//This file contains the definitions of functions to be hooked/injected when the plugin is loaded.

// How to access registers stored with PUSHAD:
// EAX => [ESP + 28]
// ECX => [ESP + 24]
// EDX => [ESP + 20]
// EBX => [ESP + 16]
// EBP => [ESP + 12]
// ESP => [ESP + 8] (warning: modifying the value of ESP is meaningless!)
// ESI => [ESP + 4]
// EDI => [ESP]

#pragma once
#include "hooks.h"
#include "definitions.h"
#include "SCBW/offsets.h"
#include "SCBW/unit.h"

#ifdef DEBUG
	#include <fstream>
	std::ofstream myfile;
#endif

unsigned char game_on = 0;

/******** Game Start ********/

void __declspec(naked) onGameStart() { // From BWAPI by Kovarex, slightly modified
	__asm PUSHAD;

	{
		game_on = 1;
		#ifdef DEBUG
			myfile << "game_on\n";
			ASSERT(gameOn());
		#else
			gameOn();
		#endif
	}
	__asm {
		POPAD
		call [VERSION::GameStartTarget]
		jmp [VERSION::GameStartBack]
	}
}


/******** Game End ********/

void __declspec(naked) onGameEnd() { // From BWAPI by Kovarex, slightly modified
	__asm PUSHAD;
	{
		game_on = 0;
		#ifdef DEBUG
			myfile << "game_end\n";
			ASSERT(gameEnd());
		#else
			gameEnd();
		#endif
	}
	__asm {
		POPAD
		call [VERSION::GameEndTarget]
		jmp [VERSION::GameEndBack]
	}
}


/******** NEXT FRAME HOOK ********/

void __declspec(naked) nextFrameHook() { // From BWAPI by Kovarex, slightly modified
	__asm {
		CALL [VERSION::NextLogicFrameTarget]
		PUSHAD
		MOV EBP, ESP
	}
	{
    #ifdef DEBUG
      myfile << "nextFrame\n";
      ASSERT(nextFrame());
    #else
      nextFrame();
    #endif
	}
	__asm {
		POPAD
		JMP [VERSION::NextLogicFrameBack]
	}
}


/******** Hooks for customizing Stim Packs HP cost ********/

//Inject with callPatch()
//Get unit pointer from ECX, store result to EDX
void __declspec(naked) stimPacksHpCostToEdx() {
	UNIT* unit;
	DWORD hpCost;
	__asm {
		PUSHAD
		MOV EBP, ESP
		MOV unit, ECX
	}

	hpCost = getStimPacksHpCost(unit);

	__asm {
		MOV EDX, hpCost
    MOV [ESP + 20], EDX   ;//Copy hpCost to stored EDX
		POPAD
    RETN
	}
}

//Inject with callPatch(); NOP 2 bytes
//Get unit pointer from ESI, compare result with unit's HP
void __declspec(naked) stimPacksHpCostCmp() {
	UNIT* unit;
	DWORD hpCost;
	__asm {
		PUSHAD
		MOV EBP, ESP
		MOV unit, ESI
	}

	hpCost = getStimPacksHpCost(unit);

	__asm {
		MOV EAX, hpCost
		CMP [unit + 8], EAX
		POPAD
    RETN
	}
}

//Inject with callPatch()
//Get unit pointer from ESI, store result to EAX
void __declspec(naked) stimPacksHpCostToEax() {
	UNIT* unit;
	DWORD hpCost;
	__asm {
		PUSHAD
		MOV EBP, ESP
		MOV unit, ESI
	}

	hpCost = getStimPacksHpCost(unit);

	__asm {
		MOV EAX, hpCost
    MOV [ESP + 28], EAX   ;//Copy hpCost to stored EAX
		POPAD
    RETN
	}
}

/******** Hooks for controlling unit HP regeneration ********/

void __declspec(naked) regenerateUnitHpHook() {
	UNIT *unit;
	__asm {
		PUSHAD
		MOV EBP, ESP
		MOV unit, ESI
	}

	regenerateUnitHp(unit);

	__asm {
		POPAD
		JMP [VERSION::RegenerateUnitHpBack]
	}
}


/******** Hook for modifying unit energy regeneration ********/

//Inject with callPatch(); NOP 2 bytes
void __declspec(naked) energyRegenAmountHook() {
	UNIT* unit;
	__asm {
		PUSHAD
		PUSHFD
		MOV EBP, ESP
		MOV unit, ESI
	}

	unit->energy += energyRegenAmount(unit);

	__asm {
		POPFD
		POPAD
		MOV DI, [ESI + 0xA2]  ;//Copy unit->energy to DI
		RETN
	}
}


/******** Hook for controlling which units can be Consumed ********/

void __declspec(naked) canConsumeUnitHook() {
	UNIT* unit;
	u8 defilerOwner;
	__asm {
		PUSHAD
		MOV EBP, ESP
		MOV unit, EAX
		MOV defilerOwner, BL
	}

	if (canConsumeUnit(unit, defilerOwner)) {
		__asm {
			POPAD
			MOV AX, 897	;//stat_txt.tbl string: Invalid target.<0>
			POP EBP
			RETN 4
		}
	}
	else {
		__asm {
			POPAD
			XOR AX, AX	;//Equivalent to MOV AX, 0 (stat_txt.tbl string: Invalid target.<0>)
			POP EBP
			RETN 4
		}
	}
}


/******** Hook for modifying how much energy is gained by Consume ********/

//Inject with jmpPatch()
void __declspec(naked) consumeEnergyGainHook() {
	UNIT* unit;
	u32 energyGain;
	__asm {
		PUSHAD
		MOV EBP, ESP
		MOV unit, ESI
	}

	energyGain = consumeEnergyGain(unit);

	__asm {
		ADD EDI, energyGain
		MOV [ESP], EDI		;//EDI <- energyGain
		POPAD
	  MOV EAX, ESI
		JMP [VERSION::ConsumeEnergyGainBack]
	}
}


/******** Hook for controlling Plague effects ********/

void __declspec(naked) doPlagueDamageHook() {
	UNIT* unit;
	__asm {
		PUSHAD
		MOV EBP, ESP
		MOV unit, ESI
	}

	doPlagueDamage(unit);

	__asm {
		POPAD
		JMP [VERSION::DoPlagueDamageBack]
	}
}


/******** Hook for controlling unit shield regeneration ********/

void __declspec(naked) regenerateShieldsHook() {
	UNIT *unit;
	__asm {
		PUSHAD
		MOV EBP, ESP
		MOV unit, ESI
	}

	if (regenerateShields(unit)) {
		__asm {
			POPAD
			JMP [VERSION::RegenerateShieldsChanged]
		}
	}
	else {
		__asm {
			POPAD
			JMP [VERSION::RegenerateShieldsUnchanged]
		}
	}
}


/******** Hooks for modifying how much shields can be recharged ********/

//Inject with callPatch(); NOP 3 bytes
//Get unit pointer from ECX, store result to EAX
void __declspec(naked) maxRechargeShields_EcxEax() {
	u32 rechargeMax;
	UNIT* unit;
	__asm {
		PUSHAD
		MOV EBP, ESP
		MOV unit, ECX
	}

	rechargeMax = maxRechargeShields(unit);

	__asm {
		MOV EAX, rechargeMax
		MOV [ESP + 28], EAX		;//EAX <- rechargeMax
		POPAD
		RETN
	}
}

//Inject with callPatch(); NOP 3 bytes
//Get unit pointer from ESI, store result to ECX
void __declspec(naked) maxRechargeShields_EsiEcx() {
	u32 rechargeMax;
	UNIT* unit;
	__asm {
		PUSHAD
		MOV EBP, ESP
		MOV unit, ESI
	}

	rechargeMax = maxRechargeShields(unit);

	__asm {
		MOV ECX, rechargeMax
		MOV [ESP + 24], ECX		;//ECX <- rechargeMax
		POPAD
		RETN
	}
}

//Inject with callPatch(); NOP 3 bytes
//Get unit pointer from EDI, store result at ECX
void __declspec(naked) maxRechargeShields_EdiEcx() {
	u32 rechargeMax;
	UNIT* unit;
	__asm {
		PUSHAD
		MOV EBP, ESP
		MOV unit, EDI
	}

	rechargeMax = maxRechargeShields(unit);

	__asm {
		MOV ECX, rechargeMax
		MOV [ESP + 24], ECX		;//ECX <- rechargeMax
		POPAD
		RETN
	}
}


/******** Hooks for modifying unit speed, acceleration, and turn speed ********/

//Inject with jmpPatch()
void __declspec(naked) modifyUnitSpeedHook() {
	UNIT *unit;
	u32 speed;
	__asm {
		PUSHAD
		MOV ebp, esp
		MOV unit, edx
		MOV speed, eax
	}

	speed = modifyUnitSpeed(unit, speed);

	__asm {
		MOV eax, speed
		MOV [esp + 28], eax
		POPAD
		RETN
	}
}

//Inject with jmpPatch()
void __declspec(naked) modifyUnitAccelerationHook() {
	UNIT *unit;
	u32 acceleration;
	__asm {
		PUSHAD
		MOV ebp, esp
		MOV unit, ecx
	}

	acceleration = modifyUnitAcceleration(unit);

	__asm {
		MOV eax, acceleration
		MOV [esp + 28], eax
		POPAD
		RETN
	}
}

//Inject with jmpPatch()
void __declspec(naked) modifyUnitTurnSpeedHook() {
	UNIT *unit;
	DWORD turnSpeed;
	__asm {
		PUSHAD
		MOV ebp, esp
		MOV unit, ecx
	}

	turnSpeed = modifyUnitTurnSpeed(unit);

	__asm {
		MOV eax, turnSpeed
		MOV [esp + 28], eax
		POPAD
		RETN
	}
}


/******** Hook for modifying weapon cooldown ********/

//Inject with jmpPatch()
void __declspec(naked) modifyWeaponCooldownHook() {
	UNIT *unit;
	u8 weaponId;
	u32 cooldown;
	__asm {
		PUSHAD
		MOV ebp, esp
		MOV unit, esi
		MOV weaponId, al
	}

	cooldown = modifyWeaponCooldown(unit, weaponId);

	__asm {
		MOV eax, cooldown
		MOV [esp + 28], eax
		POPAD
		RETN
	}
}

/******** Hook for controlling rally orders of new units ********/

void __declspec(naked) orderNewUnitToRallyHook() {
  UNIT *unit, *factory;
  __asm {
    PUSHAD
    MOV EBP, ESP
    MOV unit, ESI
    MOV factory, ECX
  }

  orderNewUnitToRally(unit, factory);

  __asm {
    POPAD
    JMP [VERSION::OrderNewUnitToRallyBack]
  }
}
