// Contains wrappers for various SCBW internal functions, to be used in
// injections.cpp.

#pragma once
#include "../definitions.h"
#include "offsets.h"
#include "unit.h"
#include "structures.h"
#include "enumerations.h"

namespace scbw {
// Prints text to the screen, optionally using the given text color.
inline void printText(char* text, DWORD color = GameTextColor::White) {
  if (!text) return;
  DWORD gtc = GetTickCount() + 7000;

  __asm   {
    PUSHAD
    PUSH 0  ;//unknown
    MOV eax, text
    PUSH gtc
    PUSH color
    CALL DWORD ptr [VERSION::PrintText]
    POPAD
  }
}

// Deals damage to the target unit.
// [attacker]: Used to increase kill count if the target dies
// [attackingPlayer]: Used to increase player scores if the target dies
// [weaponId]: Used to determine damage type (Normal, Concussive, Explosive, etc.)
// [direction]: Used to create shield damage overlays.
// [dmgDivisor]: Used for outer/medium splash damage, as well as Glave Wurm attacks.
inline void doWeaponDamage(DWORD damage,
                            UNIT* target,
                            UNIT* attacker,
                            DWORD attackingPlayer,
                            BYTE weaponId,
                            BYTE direction = 0,
                            DWORD dmgDivisor = 1) {
  if (!target || weaponId >= WEAPON_TYPE_COUNT || dmgDivisor == 0) return;

  DWORD _direction = direction;
  DWORD _weaponId = weaponId;

  __asm {
    PUSHAD
    MOV edi, target
    MOV eax, damage
    PUSH attackingPlayer
    PUSH attacker
    PUSH _direction
    PUSH dmgDivisor
    PUSH _weaponId
    CALL [VERSION::DoWeaponDamage]
    POPAD
  }
}

// Loops through the unit's image overlays, removing any that matches the given images.dat IDs.
inline void removeOverlays(UNIT *unit, DWORD imageIdStart, DWORD imageIdEnd) {
  if (!unit) return;
  if (imageIdStart > imageIdEnd) {
    DWORD tmp = imageIdStart;
    imageIdStart = imageIdEnd;
    imageIdEnd = tmp;
  }

  __asm {
    PUSHAD
    MOV eax, unit
    MOV edx, imageIdStart
    MOV edi, imageIdEnd
    CALL [VERSION::RemoveOverlays]
    POPAD
  }
}

// Loops through the unit's image overlays, removing any that matches the given images.dat ID.
inline void removeOverlays(UNIT *unit, DWORD imageId) {
  removeOverlays(unit, imageId, imageId);
}

// Creates an image overlay with the specified images.dat ID on the given sprite.
// Usage:  CreateOverlay(unit->sprite, 105);
inline void createOverlay(SPRITE* sprite, DWORD imageId, BYTE x = 0, BYTE y = 0, DWORD direction = 0) {
  if (!sprite) return;
  DWORD _x = x, _y = y;

  __asm {
    PUSHAD
    PUSH direction
    PUSH _y
    PUSH _x
    MOV esi, imageId
    MOV eax, sprite
    CALL [VERSION::CreateOverlay]
    POPAD
  }
}

// Sets the unit's HP to the specified value, and updates burning/bleeding graphics and unit strength data (used by the AI).
// If the given HP value is higher than the unit's maximum HP in units.dat, the maximum HP is used instead.
inline void setUnitHp(UNIT* unit, u32 healthPoints) {
  if (!unit) return;
  __asm {
    PUSHAD
    MOV EAX, unit
    MOV ECX, healthPoints
    CALL [VERSION::SetUnitHp]
    POPAD
  }
}

// Deals damage directly to the unit's HP, killing it if possible.
// [attackingPlayer] is only used for determining whether to deal damage when POWER OVERWHELMING is active.
// [notifyAttackedPlayer] is effective only if [attacker] points to a valid unit.
inline void damageUnitHp(UNIT* unit, u32 damage, UNIT* attacker = NULL,
                  s32 attackingPlayer = -1, u32 notifyAttackedPlayer = 1) {
  if (!unit) return;
  __asm {
    PUSHAD
    PUSH notifyAttackedPlayer
    PUSH attackingPlayer
    PUSH attacker
    MOV EAX, damage
    MOV ECX, unit
    CALL [VERSION::DamageUnitHp]
    POPAD
  }
}

/// Makes the unit use the specified weapon to attack its current target unit in
/// the UNIT::orderTarget member. This does not affect the unit's weapon
/// cooldown. The spawned weapon sprite obeys the weapon behavior properties in
/// weapons.dat.
///
/// @param  unit        The unit that uses the weapon. If NULL, the function
///                     returns without error.
/// @param  weaponId    ID of the weapon in weapons.dat. If same or bigger than
///                     WEAPON_TYPE_COUNT, the function returns without error.
inline void fireUnitWeapon(UNIT* unit, u8 weaponId) {
  if (weaponId >= WEAPON_TYPE_COUNT) return;
  u32 _weaponId = weaponId;

  __asm {
    PUSHAD
    PUSH _weaponId
    MOV ESI, unit
    CALL [VERSION::FireUnitWeapon]
    POPAD
  }
}

/// Makes the given image play the specified Iscript animation entry.
///
/// @param  image       The image to play the animation. If NULL, the function
///                     returns without error.
/// @param  animation   ID of the Iscript animation to play. See
///                     IscriptAnimation::Enum for a list of animations.
inline void playIscriptAnim(IMAGE* image, u8 animation) {
  if (!image) return;
  u32 animation_ = animation;

  __asm {
    PUSHAD
    PUSH animation_
    MOV ECX, image
    CALL [VERSION::PlayIscriptAnim]
    POPAD
  }
}

/// Makes the given unit's sprite play the specified Iscript animation entry.
///
/// @param  unit        The unit to play the animation. If NULL, the function
///                     returns without error.
/// @param  animation   ID of the Iscript animation to play. See
///                     IscriptAnimation::Enum for a list of animations.
inline void playIscriptAnim(UNIT* unit, u8 animation) {
  if (!unit || !unit->sprite) return;
  for (IMAGE* i = unit->sprite->underlay; i; i = i->next) // Based on StarCraft.exe's internal code
    playIscriptAnim(i, animation);
}

/// Attempts to create a unit at the specified coordinates. If the unit cannot
/// be created (e.g. there is no space), this function displays an error message
/// and returns NULL instead. This function is the same one used for creating
/// pre-placed units in UMS maps.
///
/// @return   The created unit or NULL if the unit cannot be created.
inline UNIT* createUnitAtPos(u16 unitType, u16 playerId, u32 x, u32 y) {
  if (unitType >= UNIT_TYPE_COUNT) return NULL;
  UNIT* unit;

  __asm {
    PUSHAD
    MOV CX, unitType
    MOV AX, playerId
    PUSH y
    PUSH x
    CALL [VERSION::CreateUnitAtPos]
    MOV [ESP + 28], EAX             ;// Overwrite the saved version of EAX
    POPAD
    MOV unit, EAX
  }

  return unit;
}

/// Issue an order to the unit. Primarily for use with orderNewUnitToRally(),
/// but may have other uses.
///
/// @param  orderId   The order ID to be used. Must be able to handle a unit as
///                   a parameter.
/// @param  target    The unit that will be used as the target. If NULL, the
///                   function returns without error.
inline void orderToUnit(UNIT* unit, u32 orderId, UNIT* target) {
  if (!target) return;

  __asm {
    PUSHAD
    PUSH orderId
    MOV EAX, target
    MOV ESI, unit
    CALL [VERSION::OrderToUnit]
    POPAD
  }
}

/// Issue an order to the unit. Primarily for use with orderNewUnitToRally(),
/// but may have other uses.
///
/// @param  orderId   The order ID to be used. Must be able to handle a unit as
///                   a parameter.
/// @param  x, y      The coordinates of the target position to use.
inline void orderToPos(UNIT* unit, u32 orderId, u16 x, u16 y) {
  u32 x_ = x, y_ = y;

  __asm {
    PUSHAD
    PUSH y_
    PUSH x_
    PUSH orderId
    MOV ESI, unit
    CALL [VERSION::OrderToPos]
    POPAD
  }
}

/// Checks whether the transport can hold this unit.
///
/// This function checks whether [transport] has any status effects (Lockdown,
/// Maelstrom, etc.), is owned by the same player, and has enough space for
/// [unit]. It does NOT check whether [transport] is a hallucination, or is an
/// Overlord without the transport upgrade, or whether the [unit] itself has any
/// status effects. Note that this function is affected by the EXE edits
/// "change bunker size check to organic flag check" and
/// "remove bunker size check" in FireGraft.
inline BOOL canBeEnteredBy(const UNIT* transport, const UNIT* unit) {
  u32 result;

  __asm {
    PUSHAD
    MOV EAX, transport
    PUSH unit
    CALL [VERSION::CanBeEnteredBy]
    MOV result, EAX
    POPAD
  }

  return result;
}

}
